//! This module defines Python bindings for statistical functions.
//! Python's GIL doesn't allow parallel computation thereof in native code, but
//! it can't stop the binding from operating pararelly on the data

use pyo3::prelude::*;
use std::sync::Arc;
use std::thread;

#[pymodule] // defines python module interface
fn data_operations(_py: Python, m: &PyModule) -> PyResult<()> {
    m.add_class::<StatisticalData>()?;
    m.add_function(wrap_pyfunction!(calculate_statistical_data, m)?)?;
    m.add_function(wrap_pyfunction!(linear_regression, m)?)?;
    Ok(())
}

#[pyclass] // this attribute defines a python class
struct StatisticalData {
    #[pyo3(get)] // allows attribute read access from python code
    length: usize,
    #[pyo3(get)]
    avg: f64,
    #[pyo3(get)]
    std_deviation: f64,
    #[pyo3(get)]
    min: f64,
    #[pyo3(get)]
    max: f64,
}

#[pyfunction] // this attribute defines a python function
fn calculate_statistical_data(data: Vec<f64>) -> Py<StatisticalData> {
    let data = Arc::new(data);

    let data_cloned = Arc::clone(&data);
    let avg_handle = thread::spawn(move || stat_fn::avg(&data_cloned));

    let data_cloned = Arc::clone(&data);
    let deviation_handle = thread::spawn(move || stat_fn::standard_deviation(&data_cloned));

    let data_cloned = Arc::clone(&data);
    let min_handle = thread::spawn(move || stat_fn::min(&data_cloned));

    let data_cloned = Arc::clone(&data);
    let max_handle = thread::spawn(move || stat_fn::max(&data_cloned));

    Python::with_gil(|py| {
        Py::new(
            py,
            StatisticalData {
                length: data.len(),
                // as this code is a simple demonstration, unwrap calls are used for simplicity
                avg: avg_handle.join().unwrap(),
                std_deviation: deviation_handle.join().unwrap(),
                min: min_handle.join().unwrap(),
                max: max_handle.join().unwrap(),
            },
        )
    })
    .unwrap()
}

#[pyfunction]
// Performs linear regression using gradient descent on a vector of 2D points
// Returns a history of weights calculated in each iteration
fn linear_regression(data: Vec<(f64, f64)>, iterations: usize, lr: f64) -> Vec<(f64, f64)> {
    let mut weights_history: Vec<(f64, f64)> = vec![];
    weights_history.reserve(iterations);

    // weights initialized to (1,1) => Ypred = x + 1
    let mut current_weights = (1.0, 1.0);

    for _ in 0..iterations {
        // closure calculating the prediction based on weights
        // basically ax + b
        let fn_pred = |x| current_weights.0 * x + current_weights.1;

        // gradients for squared error (not mean yet)
        let gradients = data
            .iter()
            .fold((0.0, 0.0), |(mut a_grad, mut b_grad), &(x, y)| {
                // common factor of both gradients
                let gradient_common = |x, y| (fn_pred(x) - y);

                a_grad += gradient_common(x, y) * x;
                b_grad += gradient_common(x, y);

                (a_grad, b_grad)
            });

        // for readability
        let len = data.len() as f64;

        current_weights.0 -= lr * 2.0 * gradients.0 / len;
        current_weights.1 -= lr * 2.0 * gradients.1 / len;

        weights_history.push(current_weights);
    }

    weights_history
}

/// Non-interface helper functions
mod stat_fn {
    pub fn avg(data: &Vec<f64>) -> f64 {
        data.iter().fold(0.0, |sum, &current| sum + current) / (data.len() as f64)
    }

    pub fn standard_deviation(data: &Vec<f64>) -> f64 {
        let avg = avg(data);
        let variation = data.iter().fold(0.0, |sum, &current| {
            sum + (current - avg).powf(2.0) / (data.len() as f64)
        });
        f64::sqrt(variation)
    }

    pub fn max(data: &Vec<f64>) -> f64 {
        let max_init = data[0];
        data.iter().fold(
            max_init,
            |max, &current| {
                if current > max {
                    current
                } else {
                    max
                }
            },
        )
    }

    pub fn min(data: &Vec<f64>) -> f64 {
        let min_init = data[0];
        data.iter().fold(
            min_init,
            |min, &current| {
                if current < min {
                    current
                } else {
                    min
                }
            },
        )
    }
}
